@@ -1,4 +1,5 @@
-//! A lightweight library that helps you detect failure of spawned async tasks without having to `.await` their handles.
+//! A lightweight library that helps you detect failure of spawned async tasks without having to
+//! `.await` their handles.
 //! Useful when you are spawning lots of detached tasks but want to fast-fail if a panic occurs.
 //!
 //! ```rust
@@ -14,15 +15,17 @@
 //!     async move {
 //!         panic!();
 //!     }
-//!     .on_panic(&alert.new_detector()) // 👈
+//!     .on_panic(&alert.new_detector()) // 👈 Binds the alert's detector
 //! );
 //!
-//! alert.drop_detector(); // See notes below
-//! assert!(alert.await.is_err());
+//! assert!(alert.drop_detector().await.is_err()); // See notes below
 //! # }
 //! ```
-//! IMPORTANT NOTE: If no task panics, the alert will hang on indefinitely, so some sort of shutdown
-//! mechanism is needed. Check out [`PanicAlert::drop_detector()`] for details.
+//! IMPORTANT NOTE: Directly `.await`ing an alert is possible, but in that case the alert as a
+//! future will only finish when a task panics. Calling `drop_detector()` allows it to finish with
+//! a `Ok(())` if no task panics as long as all the other `PanicDetector`s paired with the alert
+//! has gone out of scope. See [`PanicAlert::drop_detector`] and [`PanicMonitor::drop_detector`]
+//! for more details.
 //!
 //! For `!Send` tasks, there is the [`UnsendOnPanic`] trait:
 //! ```rust
@@ -42,8 +45,7 @@
 //!     .unsend_on_panic(&alert.new_detector())
 //! );
 //!
-//! alert.drop_detector();
-//! assert!(alert.await.is_err());
+//! assert!(alert.drop_detector().await.is_err());
 //! # }); }
 //! ```
 //!
@@ -63,21 +65,22 @@
 //! }
 //!
 //! let mut monitor = PanicMonitor::<PanicInfo>::new(); // Or simply PanicMonitor::new()
-//! let detector = monitor.new_detector();
-//!
-//! for task_id in 0..=10 {
-//!     task::spawn(
-//!         async move {
-//!             if task_id == 10 {
-//!                 panic!();
+//! {
+//!     let detector = monitor.new_detector();
+//!     for task_id in 0..=10 {
+//!         task::spawn(
+//!             async move {
+//!                 if task_id == 10 {
+//!                     panic!();
+//!                 }
 //!             }
-//!         }
-//!         // Informs the monitor of which task panicked
-//!         .on_panic_info(&detector, PanicInfo { task_id })
-//!     );
-//! }
+//!             // Informs the monitor of which task panicked
+//!             .on_panic_info(&detector, PanicInfo { task_id })
+//!         );
+//!     }
+//! } // detector goes out of scope, allowing the monitor to finish after calling drop_detector()
 //!
-//! while let Some(res) = monitor.next().await {
+//! while let Some(res) = monitor.drop_detector().next().await {
 //!     if let Err(e) = res {
 //!         let info = e.0;
 //!         assert_eq!(info.task_id, 10);
@@ -86,7 +89,6 @@
 //! }
 //! # }
 //! ```
-//!
 
 use std::{
     future::Future,
@@ -99,7 +101,6 @@ use futures::{
         mpsc::{unbounded, UnboundedReceiver, UnboundedSender},
         oneshot,
     },
-    stream::FuturesUnordered,
     FutureExt, Stream,
 };
 
@@ -152,7 +153,10 @@ where
         let (tx, rx) = oneshot::channel();
         detector
             .0
-            .unbounded_send(PanicHook { signaler: rx, info: Some(info) })
+            .unbounded_send(PanicHook {
+                signaler: rx,
+                info: Some(info),
+            })
             .expect("panic alert/monitor dropped early");
         UnsendPanicAwareFuture::new(async move {
             let ret = self.await;
@@ -221,7 +225,10 @@ where
         let (tx, rx) = oneshot::channel();
         detector
             .0
-            .unbounded_send(PanicHook { signaler: rx, info: Some(info) })
+            .unbounded_send(PanicHook {
+                signaler: rx,
+                info: Some(info),
+            })
             .expect("panic alert/monitor dropped early");
         PanicAwareFuture::new(async move {
             let ret = self.await;
@@ -278,15 +285,17 @@ where
 /// A future that finishes when a task has panicked.
 ///
 /// When one of the tasks that is being detected panics, the `PanicAlert`
-/// will finish with an `Err(Panicked)`, otherwise it will hang on indefinitely.
-/// Check out [`PanicAlert::drop_detector()`] for details.
+/// will finish with an `Err(Panicked)`, otherwise it will hang on indefinitely unless
+/// `drop_detector()` has been called and all `PanicDetector` instances have gone out of scope.
+/// Check out [`PanicAlert::drop_detector`] for details.
 pub struct PanicAlert<Info = ()>
 where
     Info: Send + 'static,
 {
     tx: Option<UnboundedSender<PanicHook<Info>>>,
     rx: UnboundedReceiver<PanicHook<Info>>,
-    list: FuturesUnordered<PanicHook<Info>>,
+    next_hook: Option<PanicHook<Info>>,
+    rx_terminated: bool,
 }
 
 impl<Info> PanicAlert<Info>
@@ -299,7 +308,8 @@ where
         PanicAlert {
             tx: Some(tx),
             rx,
-            list: FuturesUnordered::new(),
+            next_hook: None,
+            rx_terminated: false,
         }
     }
 
@@ -308,18 +318,51 @@ where
         PanicDetector(self.tx.as_ref().expect("detector already dropped").clone())
     }
 
-    /// Drops the original copy of this alert's detector so when there is no other detector instances,
-    /// the alert can finish automatically.
+    /// After calling this method, when there is no other detector instances paired with this alert,
+    /// it can finish automatically.
     ///
-    /// As an alert does not know how many tasks are to be spawned, so it will not stop working
+    /// As an alert does not know how many tasks are to be spawned, it will not stop working
     /// until:
     /// 1. A panic occurs;
-    /// 2. Or there is no `PanicDetector` paired with it.
+    /// 2. Or there is no `PanicDetector` paired with it and all tasks have finished.
     ///
-    /// Call this method will trigger the second scenario and when all the tasks bound to this
-    /// alert's detectors finish, it will finish as well.
-    pub fn drop_detector(&mut self) -> Option<PanicDetector<Info>> {
-        self.tx.take().map(|tx| PanicDetector(tx))
+    /// Calling this method will allow the second scenario to happen by dropping the `PanicDetector`
+    /// instance contained within it, so when the outstanding detectors have gone out of scope,
+    /// this alert will be able to finish.
+    ///
+    /// # Example
+    ///
+    /// The below code will park indefinitely:
+    /// ```ignore
+    /// let alert = PanicAlert::new();
+    /// task::spawn(async {}.on_panic(&alert.new_detector()));
+    /// assert!(alert.await.is_ok());
+    /// ```
+    /// But this will not:
+    /// ```no_run
+    /// # use pandet::{OnPanic, PanicAlert};
+    /// # use tokio as task;
+    /// # task::runtime::Runtime::new().unwrap().block_on(async {
+    /// let mut alert = PanicAlert::new();
+    /// task::spawn(async {}.on_panic(&alert.new_detector()));
+    /// assert!(alert.drop_detector().await.is_ok());
+    /// # });
+    /// ```
+    /// Also be careful with `PanicDetector` instances:
+    /// ```no_run
+    /// # use pandet::{OnPanic, PanicAlert};
+    /// # use tokio as task;
+    /// # task::runtime::Runtime::new().unwrap().block_on(async {
+    /// let mut alert = PanicAlert::new();
+    /// let detector = alert.new_detector();
+    /// task::spawn(async {}.on_panic(&detector));
+    /// drop(detector); // Need to drop it
+    /// assert!(alert.drop_detector().await.is_ok());
+    /// # });
+    /// ```
+    pub fn drop_detector(&mut self) -> &mut Self {
+        self.tx = None;
+        self
     }
 }
 
@@ -330,44 +373,50 @@ where
     type Output = Result<(), Panicked<Info>>;
 
     fn poll(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
-        let rx_terminated = match Pin::new(&mut self.rx).poll_next(cx) {
-            Poll::Pending => false,
-            Poll::Ready(Some(r)) => {
-                self.list.push(r);
-                false
+        Poll::Ready(loop {
+            if self.next_hook.is_none() && !self.rx_terminated {
+                self.rx_terminated = match Pin::new(&mut self.rx).poll_next(cx) {
+                    Poll::Pending => return Poll::Pending,
+                    Poll::Ready(Some(r)) => {
+                        self.next_hook = Some(r);
+                        false
+                    }
+                    Poll::Ready(None) => true,
+                };
             }
-            Poll::Ready(None) => true,
-        };
-        match Pin::new(&mut self.list).poll_next(cx) {
-            Poll::Ready(Some(r)) => {
-                if r.is_err() {
-                    Poll::Ready(r)
-                } else {
-                    Poll::Pending
+
+            if let Some(mut next) = self.next_hook.take() {
+                let res = Pin::new(&mut next).poll(cx);
+                match res {
+                    Poll::Ready(r) => {
+                        if r.is_err() {
+                            break r;
+                        }
+                    }
+                    Poll::Pending => {
+                        self.next_hook = Some(next);
+                        return Poll::Pending;
+                    }
                 }
+            } else {
+                break Ok(());
             }
-            Poll::Ready(None) => {
-                if rx_terminated {
-                    Poll::Ready(Ok(()))
-                } else {
-                    Poll::Pending
-                }
-            }
-            Poll::Pending => Poll::Pending,
-        }
+        })
     }
 }
 
 /// A [`Stream`](https://docs.rs/futures/latest/futures/stream/trait.Stream.html#) of detected panics.
 ///
-/// Check out [`PanicMonitor::drop_detector()`] for details about when this stream will finish.
+/// This stream emits `Some(Err(Panicked))` when a tasks panics, and nothing otherwise.
+/// Check out [`PanicMonitor::drop_detector`] for details about when this stream will finish.
 pub struct PanicMonitor<Info = ()>
 where
     Info: Send + 'static,
 {
     tx: Option<UnboundedSender<PanicHook<Info>>>,
     rx: UnboundedReceiver<PanicHook<Info>>,
-    list: FuturesUnordered<PanicHook<Info>>,
+    next_hook: Option<PanicHook<Info>>,
+    rx_terminated: bool,
 }
 
 impl<Info> PanicMonitor<Info>
@@ -378,9 +427,10 @@ where
     pub fn new() -> Self {
         let (tx, rx) = unbounded();
         PanicMonitor {
-            tx : Some(tx),
+            tx: Some(tx),
             rx,
-            list: FuturesUnordered::new(),
+            next_hook: None,
+            rx_terminated: false,
         }
     }
 
@@ -389,18 +439,50 @@ where
         PanicDetector(self.tx.as_ref().expect("detector already dropped").clone())
     }
 
-    /// Drops the original copy of this monitor's detector so when there is no other detector instances,
-    /// the monitor can finish automatically.
+    /// After calling this method, when there is no other detector instances paired with this monitor,
+    /// it can finish automatically.
     ///
-    /// As a monitor does not know how many tasks are to be spawned, so it will not stop working
-    /// until:
-    /// 1. The caller actively stops polling the stream;
-    /// 2. Or there is no `PanicDetector` paired with it.
+    /// As a monitor does not know how many tasks are to be spawned, it will not stop working
+    /// until there is no `PanicDetector` paired with it and all tasks have finished.
     ///
-    /// Call this method will trigger the second scenario and when all the tasks bound to this
-    /// monitor's detectors finish, it will finish as well.
-    pub fn drop_detector(&mut self) -> Option<PanicDetector<Info>> {
-        self.tx.take().map(|tx| PanicDetector(tx))
+    /// Calling this method will drop the `PanicDetector` instance contained within it, so when the
+    /// outstanding detectors have gone out of scope, this monitor will be able to finish.
+    ///
+    /// # Example
+    ///
+    /// The below code will park indefinitely:
+    /// ```ignore
+    /// let mut monitor = PanicMonitor::new();
+    /// task::spawn(async {}.on_panic(&monitor.new_detector()));
+    /// while let Some(_) = monitor.next().await {}
+    /// ```
+    /// But this will not:
+    /// ```no_run
+    /// # use pandet::{OnPanic, PanicMonitor};
+    /// # use tokio as task;
+    /// # use futures::StreamExt;
+    /// # task::runtime::Runtime::new().unwrap().block_on(async {
+    /// let mut monitor = PanicMonitor::new();
+    /// task::spawn(async {}.on_panic(&monitor.new_detector()));
+    /// while let Some(_) = monitor.drop_detector().next().await {}
+    /// # });
+    /// ```
+    /// Also be careful with `PanicDetector` instances:
+    /// ```no_run
+    /// # use pandet::{OnPanic, PanicMonitor};
+    /// # use tokio as task;
+    /// # use futures::StreamExt;
+    /// # task::runtime::Runtime::new().unwrap().block_on(async {
+    /// let mut monitor = PanicMonitor::new();
+    /// let detector = monitor.new_detector();
+    /// task::spawn(async {}.on_panic(&detector));
+    /// drop(detector); // Need to drop it
+    /// while let Some(_) = monitor.drop_detector().next().await {}
+    /// # });
+    /// ```
+    pub fn drop_detector(&mut self) -> &mut Self {
+        self.tx = None;
+        self
     }
 }
 
@@ -411,25 +493,35 @@ where
     type Item = Result<(), Panicked<Info>>;
 
     fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
-        let rx_terminated = match Pin::new(&mut self.rx).poll_next(cx) {
-            Poll::Pending => false,
-            Poll::Ready(Some(r)) => {
-                self.list.push(r);
-                false
+        Poll::Ready(loop {
+            if self.next_hook.is_none() && !self.rx_terminated {
+                self.rx_terminated = match Pin::new(&mut self.rx).poll_next(cx) {
+                    Poll::Pending => return Poll::Pending,
+                    Poll::Ready(Some(r)) => {
+                        self.next_hook = Some(r);
+                        false
+                    }
+                    Poll::Ready(None) => true,
+                };
             }
-            Poll::Ready(None) => true,
-        };
-        match Pin::new(&mut self.list).poll_next(cx) {
-            Poll::Ready(Some(r)) => Poll::Ready(Some(r)),
-            Poll::Ready(None) => {
-                if rx_terminated {
-                    Poll::Ready(None)
-                } else {
-                    Poll::Pending
+
+            if let Some(mut next) = self.next_hook.take() {
+                let res = Pin::new(&mut next).poll(cx);
+                match res {
+                    Poll::Ready(r) => {
+                        if r.is_err() {
+                            break Some(r);
+                        }
+                    }
+                    Poll::Pending => {
+                        self.next_hook = Some(next);
+                        return Poll::Pending;
+                    }
                 }
+            } else {
+                break None;
             }
-            Poll::Pending => Poll::Pending,
-        }
+        })
     }
 }
 
@@ -446,14 +538,24 @@ mod tests {
         let alert = PanicAlert::new();
         let detector = alert.new_detector();
 
-        tokio::spawn(
-            async move {
-                panic!();
-            }
-            .on_panic(&detector),
-        );
-
+        for i in 0..=10 {
+            tokio::spawn(
+                async move {
+                    if i == 1 {
+                        panic!("What could go wrong");
+                    }
+                }
+                .on_panic(&detector),
+            );
+        }
         assert!(alert.await.is_err());
+
+        let mut alert = PanicAlert::new();
+        (0..=10).for_each(|_| {
+            let detector = alert.new_detector();
+            tokio::spawn((move ||{async move {}.on_panic(&detector)})());
+        });
+        assert!(alert.drop_detector().await.is_ok());
     }
 
     #[tokio::test]
@@ -479,24 +581,22 @@ mod tests {
     #[tokio::test]
     async fn monitor_works() {
         let mut monitor = PanicMonitor::new();
-        let detector = monitor.new_detector();
 
         for i in 0..=10 {
             tokio::spawn(
                 async move {
-                    if i == 10 {
+                    if i == 5 {
                         panic!();
                     }
                 }
-                .on_panic_info(&detector, 10),
+                .on_panic_info(&monitor.new_detector(), i),
             );
         }
 
-        while let Some(res) = monitor.next().await {
+        while let Some(res) = monitor.drop_detector().next().await {
             if res.is_err() {
                 let id = res.unwrap_err().0;
-                assert_eq!(id, 10);
-                break;
+                assert_eq!(id, 5);
             }
         }
     }
